
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"
#include "cbp.h"
#include "sif_f.h"

extern void	Error(char *msg);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error("multi-threading initialization error.")
#define TERM_MT(T)	\
if (TERM_THREAD(T)) Error("multi-threading termination error.")

#define LEN	2048

typedef struct {
		int	Nx,Ny;
		FOM	**fom;
		char	path[LEN],info[LEN];
	} Struct;

static FUNCTION_T	Store(void *a)
{
	Struct	*s=(Struct *)a;

	StoreImageFile_Float(s->path,s->Nx,s->Ny,s->fom,SIF_F_desc);
	(void)printf("%s\t%s\n",s->info,SIF_F_desc);

	return RETURN_VALUE;
}

#define EPS	1e-9
#define DEG	3.14159265358979323846/180.0

int	main(int argc,char **argv)
{
	HiPic		hp;
	double		Dr,base,step,RA0,y0,dy,RC,Dz=0.0;
	int		count,Nz,z,x,y,t,len,c;
	Float		**P,*P_F,**F;
	Struct		S;
	FOM		*S_fom,*hp_T,fom;
	char		form[LEN];
	THREAD_T	T;

	if (argc!=9 && argc!=10)
    Error("usage : hp_stg HiPic/ Dr RC_base step count {Dz} layer RA0 STG/");

	InitReadHiPic(argv[1],&hp);

	if (hp.Nt<2) Error("too few views.");

	if ((Dr=atof(argv[2]))<EPS)
	    Error("bad horizontal interval of detectors on HiPic image.");

	if ((base=atof(argv[3]))<=0.0 ||
	    (step=atof(argv[4]))<=0.0 ||
	    (count=atoi(argv[5]))<=0 ||
	    base+step*(double)(count-1)>=(double)(hp.Nx-1))
	    Error("bad range of rotation center.");

	if (argc==9)
	    Nz=hp.Ny;
	else
	    if ((Nz=hp.Ny+(int)ceil(fabs(Dz=atof(argv[6]))))>=2*hp.Ny)
		Error("too large vertical displacement.");

	if ((z=atoi(argv[argc-3]))<0 || z>=Nz) Error("bad slice position.");

	RA0=atof(argv[argc-2])*DEG;

	P=InitCBP(hp.Nx,hp.Nt);

	x=(hp.Nx>hp.Nt)?hp.Nx:hp.Nt;
	if ((S.fom=(FOM **)malloc(sizeof(FOM *)*x))==NULL ||
	    (*(S.fom)=(FOM *)malloc(sizeof(FOM)*x*hp.Nx))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<x; y++) S.fom[y]=S.fom[y-1]+hp.Nx;

	y0=0.5+(double)z; if (Dz<0.0) y0+=Dz; dy=Dz/(double)hp.Nt;
	for (t=0; t<hp.Nt; t++) {
	    S_fom=S.fom[t]; P_F=P[t];
	    if ((y=(int)floor(y0-dy*(double)t))<0 || y>=hp.Ny)
		for (x=0; x<hp.Nx; x++) *(S_fom++)=(*(P_F++))=0.0;
	    else {
		ReadHiPic(&hp,t);

		hp_T=hp.T[y];
		for (x=0; x<hp.Nx; x++)
		    *(S_fom++)=(*(P_F++))=((fom=(*(hp_T++)))>0.0)?-log(fom):0.0;
	    }
	}
	TermReadHiPic(&hp);

	len=1; for (y=10; y<Nz; y*=10) ++len;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],len);
	(void)sprintf(S.path,form,z);
	(void)sprintf(S.info,"%d",z);

	S.Nx=hp.Nx; S.Ny=hp.Nt;

	INIT_MT(T,Store,&S);

	c=1+2; for (x=10; x<hp.Nx-1; x*=10) ++c;
	(void)sprintf(form,"%s/%%0%dd_%%0%d.1lf.tif",argv[argc-1],len,c);

	for (c=0; c<count; c++) {
	    F=CBP(Dr,-(RC=base+step*(double)c),RA0);

	    TERM_MT(T);

	    if (c==0) S.Ny=hp.Nx;

	    (void)sprintf(S.path,form,z,RC);
	    (void)sprintf(S.info,"%d\t%lf",z,RC);

	    for (y=0; y<hp.Nx; y++) {
		S_fom=S.fom[y]; P_F=F[y];
		for (x=0; x<hp.Nx; x++) *(S_fom++)=(*(P_F++));
	    }
	    INIT_MT(T,Store,&S);
	}
	TERM_MT(T);

	free(*(S.fom)); free(S.fom); TermCBP();

	return 0;
}
