
#include <stdio.h>
#include <stdlib.h>
#include "csi_f.h"
#include "rif_f.h"
#include "sif_f.h"

static int	x1,y1,x2,y2;
static FOM	**Fn;

#define LEN	2048

static void	Trim_Float(char *Po,char *Pn)
{
	int	bps,Nx,Ny,n,y,x;
	FOM	**Fo,*fom;
	char	*id,str[8][LEN],*u_c,
		*u="using integer PV as is floating-point PV",
		*c="converting PV by TIFF image description tag";
	double	min,max,b,s;

	if ((bps=ReadImageFile_Float(Po,&Nx,&Ny,&Fo,&id))!=0) {
	    if (id!=NULL) {
		if ((n=sscanf(id,"%s %s %s %s %s %s %s %s",
				 str[0],str[1],str[2],str[3],
				 str[4],str[5],str[6],str[7]))==2 || n==6)
		    if (sscanf(str[n-2],"%lf",&min)!=1 ||
			sscanf(str[n-1],"%lf",&max)!=1 || min>=max)
			u_c=u;
		    else {
			u_c=c; b=min; s=(max-min)/(double)((1ULL<<bps)-1);
		    }
		else
		    u_c=((n!=3 && n!=4 && n!=8) ||
			 sscanf(str[n-2],"%lf",&b)!=1 ||
			 sscanf(str[n-1],"%lf",&s)!=1 || s<=0.0)?u:c;

		free(id);
	    }
	    else
		u_c=u;
#ifdef	U_C
	    (void)fprintf(stderr,"%s : %s (warning).\n",Po,u_c);
#endif
	}
	else {
	    u_c=NULL; if (id!=NULL) free(id);
	}
	if (u_c!=c) {
	    b=0.0; s=1.0;
	}
	fom=(*Fn);
	for (y=y1; y<=y2 && y<Ny; y++) {
	    for (x=x1; x<=x2 && x<Nx; x++) *(fom++)=b+s*Fo[y][x];

	    for (; x<=x2; x++) *(fom++)=b;
	}
	for (; y<=y2; y++)
	    for (x=x1; x<=x2; x++) (*fom)=b;

	for (y=0; y<Ny; y++) free(Fo[y]); free(Fo);

	StoreImageFile_Float(Pn,x2-x1+1,y2-y1+1,Fn,SIF_F_desc);
}

int	main(int argc,char **argv)
{
	int	Nx,Ny,Nz,z1,z2,z,len;
	char	**Po,Pn[LEN],form[LEN];

	if (argc==7) {
	    (void)ReadImageFile_Float(argv[1],&Nx,&Ny,NULL,NULL); Nz=1;

	    x1=atoi(argv[2]); y1=atoi(argv[3]);
	    x2=atoi(argv[4]); y2=atoi(argv[5]); z1=z2=0;
	}
	else if (argc==8 || argc==10) {
	    Po=CheckSliceImages_Float(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	    x1=atoi(argv[3]); y1=atoi(argv[4]);
	    if (argc==8) {
		x2=atoi(argv[5]); y2=atoi(argv[6]); z1=0; z2=Nz-1;
	    }
	    else {
		z1=atoi(argv[5]); x2=atoi(argv[6]);
				  y2=atoi(argv[7]); z2=atoi(argv[8]);
	    }
	}
	else {
(void)fputs("usage : trim_float org.tif x1 y1 x2 y2 new.tif\n"
	    "        trim_float org/ nameFile x1 y1 x2 y2 new/\n"
	    "        trim_float org/ nameFile x1 y1 z1 x2 y2 z2 new/\n",stderr);
	    exit(1);
	}
	if (x1<0 || x2>=Nx || x1>x2 ||
	    y1<0 || y2>=Ny || y1>y2 ||
	    z1<0 || z2>=Nz || z1>z2) {
	    (void)fputs("bad triming area.\n",stderr); exit(1);
	}
	if ((Fn=(FOM **)malloc(sizeof(FOM *)*(Ny=y2-y1+1)))==NULL ||
	    (*Fn=(FOM *)malloc(sizeof(FOM)*Ny*(Nx=x2-x1+1)))==NULL) {
	    (void)fputs("memory allocation error.\n",stderr); exit(1);
	}
	for (z=1; z<Ny; z++) Fn[z]=Fn[z-1]+Nx;

	if (argc==7)
	    Trim_Float(argv[1],argv[6]);
	else {
	    if (argc==8) {
		for (len=0; argv[1][len]!='\0'; len++) ;

		for (z=0; z<Nz; z++) {
		    (void)sprintf(Pn,"%s/%s",argv[7],Po[z]+len+1);
		    Trim_Float(Po[z],Pn);
		}
	    }
	    else {
		len=1; for (z=10; z<z2-z1+1; z*=10) ++len;
		(void)sprintf(form,"%s/%%0%dd.tif",argv[9],len);

		for (z=z1; z<=z2; z++) {
		    (void)sprintf(Pn,form,z-z1);
		    Trim_Float(Po[z],Pn);
		}
	    }
	    for (z=0; z<Nz; z++) free(Po[z]); free(Po);
	}
	free(*Fn); free(Fn);

	return 0;
}
